# -*- coding: utf-8 -*-
"""
Created on Thu May  7 12:14:04 2020
@author: Sondr
##"""
#
import numpy as np
import matplotlib.pyplot as plt
import numpy.linalg as la
#from control import *
from casadi import *
from PNMPC_LIB_MOD import *
import cvxopt
import scipy.io as sio
import mat4py as m4p
#this is a small crash course on how to run PNMPC control....

print("If the error IDA_LINESEARCH_FAIL, run the program more times. It comes from the model gets negative values")
#
##first, we need a system, so I created a discrete linear model as a toy problem.
class linear_discrete(SimulationModel):
    
    def __init__(self):
        #MODEL CONSTANTS:
        g = 9.81 #Gravity
        Cc = 2*10**(-5) #Chove valve constant
        A1 = 0.008107 #Cross section below ESP
        A2 = 0.008107 #Cross section above ESP
        D1 = 0.1016 #Pipe diameter below ESP
        D2 = 0.1016 #Pipe diameter above ESP
        h1 = 200 #Height from reservoir to ESP
        h2 = 800 #Height ESP up
        hw = 1000 #Total distance in well
        L1 = 500 #Length reservoir to ESP
        L2 = 1200 #Length from ESP to choce
        V1 = 4.054 #Pipe volume below ESP
        V2 = 9.729 #Pipe volum above ESP
        #ESP DATA 
        B1 = 1.5*10**9 #Bulk modulus below ESP
        B2 = 1.5*10**9 #Bulk modulus above ESP
        M = 1.992*10**8 #Fluid inertia
        rho = 950 #Density of fluid
        pr = 1.26 * 10 ** 7 #Reservoir pressure
        f0 = 60 # ESP reference freq
        #UNKNOWN
        PI = 2.32*10**(-9) #Well producity index
        pm = 20e5 #Manifold pressure
        my = 0.025 # Vicosity of produced fluid
        
        # VCF for ESP flow rate
        CQ = 1 - 2.6266*my + 6.0032*my**2 - 6.8104*my**3 + 2.7944*my**4
        CH = 1 - 0.03*my
        
        #INITIAL VALUES
        self.pbh0 = 70e5
        self.q0 = 0.01
        self.pwh0 = 30e5
        
        #END CONSTANTS
        
        #INITIALIZE MODEL
        dae = DaeBuilder()
        # Add input expressions
        pbh = dae.add_x('pbh')
        q = dae.add_x('q')
        pwh = dae.add_x('pwh')
        
        
        qr = dae.add_z('qr')
        qc = dae.add_z('qc')
        
        F1 = dae.add_z('F1')
        F2 = dae.add_z('F2')
        DeltaPp = dae.add_z('DeltaPp')
        
        u = dae.add_p('u')
        f = dae.add_p('f')
        
        # Add output expressions
        pbhdot = (B1/V1)*(qr-q)
        qdot = (1/M)*(pbh-pwh-rho*g*hw - F1 - F2 + DeltaPp)
        pwhdot = (B2/V2)*(q-qc)
        
        
        dae.add_ode('pbhdot', pbhdot)
        dae.add_ode('qdot', qdot)
        dae.add_ode('pwhdot', pwhdot)
        
        
        #Algebraic equation
        dae.add_alg('qr', qr-PI*(pr-pbh))
        dae.add_alg('qc', qc-Cc*(sqrt(fmax(10**-5, pwh-pm))*u))
        
        dae.add_alg('F1', F1 - 0.158*((rho*L1*q**2)/(D1*A1**2))*(my/(rho*D1*q))**(1/4))
        dae.add_alg('F2', F2 - 0.158*((rho*L2*q**2)/(D2*A2**2))*(my/(rho*D2*q))**(1/4))
        
        # Algebraic ESP equations:
        dae.add_alg('DeltaPp', DeltaPp - rho*g* (CH*(9.5970e2 + 7.4959e3*((q/CQ)*(f0/f)) - 1.2454e6*((q/CQ)*(f0/f))**2)*(f/f0)**2))
        #CREATE DAE
        dae = {'x': vertcat(*dae.x), 'p': vertcat(*dae.p),'z': vertcat(*dae.z), 'ode': vertcat(*dae.ode), 'alg': vertcat(*dae.alg)}
        ts = np.linspace(0,0.05,2)

        opts = {'grid':ts}
        self.I = integrator('I', 'idas', dae, opts)
        

        
    	#START MED INITAL, så gå over til first result
    def model_output(self,u):
        I = self.I
#        print(u[1])
        z = u[0]
        f = u[1]
      #FIRST STEP WITH INPUT z =1 , f = 40
        x0 = vertcat(self.pbh0,self.q0,self.pwh0) #DIFFERENTIAL
#        print(x0)
        p0 = vertcat(z,f) #INPUTS
        res = self.I(x0=x0,p=p0)
        solution = res['xf'].full().T
#        print(solution)
        self.pbh0 = solution[len(solution)-1,0] #Initalizing model with correct states with new input
        self.pwh0 = solution[len(solution)-1,2]
        self.q0 = fmax(1e-5,solution[len(solution)-1,1])
        y = solution[len(solution)-1,:]
        return y
    
    

ESP = linear_discrete()
#------------------------------------------



#we first need to gather data, so we use the built-in RFRAS (Random Frequency Random Amplitude Signal)


#these lines of code create the identificaton signal:
simtime = 6000
u_min = np.array([0.1])
u_max = np.array([1])
some_input_u_1 = RFRAS(u_min,u_max,simtime,10)
some_input_u = RFRAS(u_min,u_max,simtime,40) #30

f_min = np.array([35])
f_max = np.array([65])
some_input_f_1 = RFRAS(f_min,f_max,simtime,10)
some_input_f = RFRAS(f_min,f_max,simtime,40)
#print(some_input)
#THe data MUST be 2dimensional, and data must organized in row form, so in this case:

some_input_u = np.atleast_2d(some_input_u).T
some_input_f = np.atleast_2d(some_input_f).T

some_input_u_1 = np.atleast_2d(some_input_u_1).T
some_input_f_1 = np.atleast_2d(some_input_f_1).T

some_input = np.zeros([12000,2])

for i in range (0,len(some_input_u_1)):
    some_input[i,0] = some_input_u_1[i]
    some_input[i,1] = some_input_f_1[i]
    
    
print(len(some_input_u_1))

#for i in range (0,len(some_input_u_1)):
#    some_input[i+6000,0] = some_input_u[i]
#    some_input[i+6000,1] = some_input_f[i]

for i in range (0,len(some_input_u_1)):
    some_input[i+6000,0] = some_input_u[i]
    some_input[i+6000,1] = some_input_f[i]

#some_input = [*some_input_u , *some_input_f]

#the resulting dimension should always be (number of data x number of features)

#now we need to run the system with that signal to obtain the system response, which is recorded in y_data
y_data = np.empty([12000,3])
y_data[0] = (70e5,0.01,30e5)
#the simulation and plotting for data gathering is as follows:
for i in range(len(some_input)):
    y = ESP.model_output(some_input[i])
    y_data[i] = y
#print(y_data)


#PLOTS OF THE DATA INCLUDED THE INPUTS
plt.figure(1)
plt.rcParams.update({'font.size': 22})
plt.subplot(511)
plt.plot(y_data[:,0]/10**5,color='blue', label='$p_{bh}$',linewidth=2.0)
plt.grid()
plt.rc('ytick', labelsize=20)
plt.rc('xtick', labelsize=20)  
plt.legend()
plt.ylabel('$p_{bh} \ [bar]$')


plt.subplot(512)
plt.plot(y_data[:,1]*3600,color='blue',label='$q$',linewidth=2.0)
plt.grid()
plt.legend()
plt.rc('xtick', labelsize=20)  
plt.rc('ytick', labelsize=20) 
plt.ylabel('$p_{wh} \ [bar]$')

plt.subplot(513)
plt.plot(y_data[:,2]/10**5,color='blue',label='$p_{wh}$',linewidth=2.0)
plt.rc('ytick', labelsize=20) 
plt.rc('xtick', labelsize=20)  
plt.grid()
plt.legend()
plt.ylabel(r'$q \ [\dfrac{m^3}{h}]$')
#plt.xlabel('Time [min]')

plt.subplot(514)
plt.plot(some_input[:,0],color='red',label='$valve$',linewidth=2.0)
plt.rc('ytick', labelsize=20) 
plt.rc('xtick', labelsize=20)  
plt.grid()
plt.legend()
plt.ylabel('$\%$')
#plt.xlabel('Time [min]')

plt.subplot(515)
plt.plot(some_input[:,1],color='red',label='$Frequency$',linewidth=2.0)
plt.rc('ytick', labelsize=20) 
plt.rc('xtick', labelsize=20)  
plt.grid()
plt.legend()
plt.ylabel('$Hz$')
plt.xlabel('Time [min]')


#m4p.savemat('datafile.mat', y_data[0,:])
#
#sio.savemat('y_data.mat',y_data[:,0])
#sio.savemat('some_input.mat',some_input)

#
#####plt.plot(y_data)
###------------------------------------------------------------


#next step is to normalize the data using feature scaling, so we need to use
#a minimum value and a maximum value, even better if it ties to the constraints of your problem.
#
##The scaling is done as follows, using the built_in function from esn_pnmpc_library:
u_data_normal = feature_scaling(some_input,u_max,u_min)
y_max = np.max(y_data,axis=0)
y_min = np.min(y_data,axis=0)
y_data_normal = feature_scaling(y_data,y_max,y_min)


u_data_normal = np.zeros([len(some_input),2])
y_data_normal = np.zeros([len(some_input),2])

u_data_normal[:,0] = feature_scaling(some_input[:,0],u_max,u_min)
u_data_normal[:,1] = feature_scaling(some_input[:,1],f_max,f_min)


#y_max = np.max(y_data,axis=0)
#y_min = np.min(y_data,axis=0)
#y_data_normal = feature_scaling(y_data,y_max,y_min)
#

##---------------------------------------------------------------------------------------------------
#
##Now, ater obtaining a set of normalized inputs and outputs, we shall instantiate and train a black-box model:
#
##in this case, an ESN is used, with the parameters arbitrariely selected.
##Obviously, the ESN is a bit overkill for the problem of training a linear third order system...
##

toy_esn = EchoStateNetwork(neu = 400,
                           n_in = 2,
                           n_out = 3,
                 gama=0.14,
                 ro=0.99,
                 psi=0.0,
                 in_scale=0.1,
                 bias_scale=0.1,
                 initial_filename="initial",
                 load_initial = False,
                 save_initial = False)
#                 noise_amplitude = 5e-3)
#toy_esn = RNN.EchoStateNetwork(neu = 100,n_in = 2,n_out = 3,gama = 0.14,ro = 0.99,psi = 0.0,in_scale = 0.1,bias_scale = 0.1, initial_filename = "initial", load_initial = True, save_initial = False,output_feedback = False)
#-----------------------------------------------------------------------------------------------


#now, we need to train it, by first feeding the ESN with the normalized data.
#the warmupdrop means that the first few [warmupdrop] examples will the dropped.
#In this case, the first few examples are outliers and have a high probability of corrupting the data.

#the function add data serves to gather data of one simulation run to the class "database".
#Multiples simulation could also be absorbed by multiple runs of the command.

#now, we train. The training function also returns the best regularization parameter while insternally changing the weights.
y_data_normal = feature_scaling(y_data,y_max,y_min)
regularization = 1e-8	 
drop = 200

warmupdrop = 200

#offline_training(self,X,Y,regularization,warmupdrop):
toy_esn.offline_training(u_data_normal,y_data_normal,regularization,warmupdrop)

#print("TEST")

#error, reg = toy_esn.cum_train_cv(min_reg,max_reg)
                 

 	 
#adter training the ESN, it is recommended you generate another input signal to test the ESN:

simtime = 300

#
#CREATING TEST INPUT WITH RANDOM IN
test_input_u = RFRAS(u_min,u_max,simtime,10)
test_input_f = RFRAS(f_min,f_max,simtime,10)

test_input = np.zeros([len(test_input_u),2])

for i in range (len(test_input_u)):
    test_input[i,0] = test_input_u[i]
    test_input[i,1] = test_input_f[i]

#CREATING THE PREDICTION WITH ESN 
y_pred_test = np.empty([simtime,3])
y_test = np.empty_like(y_pred_test)

for k in range(simtime):
    y_test[k] = ESP.model_output(test_input[k])


test_input[:,0] = feature_scaling(test_input[:,0],u_max,u_min)
test_input[:,1] = feature_scaling(test_input[:,1],f_max,f_min)


for k in range(simtime):
    y_pred_test[k] = toy_esn.update(test_input[k]).flatten()
    y_pred_test[k] = feature_descaling(y_pred_test[k],y_max,y_min)


#PLOTS WITH REAL AND PREDICTED VALUES
plt.figure(2)
plt.subplot(511)
plt.plot(y_test[:,0]/10**5,color='blue', label='$p_{bh}$')
plt.plot(y_pred_test[:,0]/10**5,color='orange', label='$pred$')
plt.grid()
plt.legend()
plt.ylabel('$p_{bh} \ [bar]$')

plt.subplot(512)
plt.plot(y_test[:,1]/10**5,color='blue', label='$p_{wh}$')
plt.plot(y_pred_test[:,1]/10**5,color='orange', label='$pred$')
plt.grid()
plt.legend()
plt.ylabel('$p_{wh} \ [bar]$')

plt.subplot(513)
plt.plot(y_test[:,2]*3600,color='blue', label='$q$')
plt.plot(y_pred_test[:,2]*3600,color='orange', label='$pred$')
plt.grid()
plt.legend()
plt.ylabel(r'$q \ [\dfrac{m^3}{h}]$')

plt.subplot(514)
plt.plot(test_input[:,0],color='red',label='$valve$')
plt.grid()
plt.legend()

plt.subplot(515)
plt.plot(test_input[:,1],color='red',label='$frekvens$')
plt.grid()
plt.legend()
plt.xlabel('Time [min]')
#from the plot it seems the esn can easily copy the system behavior, therefore we are ready to instantiate the predictor:


some_input = [[1,65],
              [0.1,35]]

u_max = np.max(some_input,axis=0)
u_min = np.min(some_input,axis=0)



a = 0.5 #forcing the pole of the filter.

toy_predictor = PNMPCPredictor(control_horizon = 10,
                               prediction_horizon = 30,
                               bb_model = toy_esn,
                               filter_constants = a**2, 
                               K = (a**2 - 2.0*a +1)/(1 - a**2))

#now that we just created the predictor, we need to create the control instance....



#initial_u = [[0.7],[40]]
initial_u = [0.75,49] #0.65 , 6
initial_u = np.atleast_1d(initial_u)
print("test",initial_u)
control = PNMPCController(n_cv = 2,
                 u_max = u_max,
                 u_min = u_min,
                 y_max = y_max,
                 y_min = y_min, 
                 initial_u = initial_u,
                 Q = np.array([[0.07, 0], [0, 0.001]]), #0.1, 0.001
                 R = np.array([[3, 0], [0, 5]]), #3,5
                 du_max_c = 10,
                 output_constraints_max = 30e5,
                 output_constraints_min = 0,
                 predictor = toy_predictor,
                 simulation_model = ESP)


# R = np.eye(2)*10,

#running the controller is simple, you just have to define a proper reference signal for it, of dimension n_cv
pnmpc_time = 720    
#y_ref = RFRAS(y_min[0],y_max[0],pnmpc_time,20)
#reference should be organized just like the data matrices, this is needed because there is only one controlled variable:
#y_ref = np.atleast_2d(y_ref).T

y_ref = np.zeros([pnmpc_time,2])

#for i in range(pnmpc_time):
#    y_ref[i,0] = 70e5
#    y_ref[i,1] = 25e5
#    y_ref[i,2] = 0.007
#for i in range(250,pnmpc_time):
#    y_ref[i,0] = 60e5
#    y_ref[i,1] = 60e5
#print(y_ref.shape[0])

######################
#for i in range(360):
#    y_ref[i,0] = 82e5
#    y_ref[i,1] = 0.018
#
#for i in range(360,720):
#    y_ref[i,0] = 70e5
#    y_ref[i,1] = 0.018

###################
#for i in range(180):
#    y_ref[i,0] = 82e5
#    y_ref[i,1] = 0.018
#
#for i in range(180,360):
#    y_ref[i,0] = 79e5
#    y_ref[i,1] = 0.018
#    
#for i in range(360,540):
#    y_ref[i,0] = 76e5
#    y_ref[i,1] = 0.018
#    
#for i in range(540,720):
#    y_ref[i,0] = 73e5
#    y_ref[i,1] = 0.018
#################

for i in range(90):
    y_ref[i,0] = 82e5
    y_ref[i,1] = 0.018

for i in range(90,180):
    y_ref[i,0] = 79e5
    y_ref[i,1] = 0.018
    
for i in range(180,270):
    y_ref[i,0] = 76e5
    y_ref[i,1] = 0.018
    
for i in range(270,360):
    y_ref[i,0] = 73e5
    y_ref[i,1] = 0.018
    
for i in range(360, 450):
    y_ref[i,0] = 73e5
    y_ref[i,1] = 0.018

for i in range(450,540):
    y_ref[i,0] = 76e5
    y_ref[i,1] = 0.018
    
for i in range(540,630):
    y_ref[i,0] = 79e5
    y_ref[i,1] = 0.018
    
for i in range(630,720):
    y_ref[i,0] = 82e5
    y_ref[i,1] = 0.018
    

    
    
#y_ref = np.zeros([pnmpc_time,1])
#for i in range(pnmpc_time):
#    y_ref[i] = 75e5

    
#for i in range(100,pnmpc_time):
#    y_ref[i,0] = 60e5
#    y_ref[i,1] = 30e5
    
#for i in range(200,pnmpc_time):
#    y_ref[i,0] = 80e5
#    y_ref[i,1] = 22e5


#y_ref = np.array([[92e5],[25e5],[0.007]])
#y_ref = np.array([[92e5],[25e5]])
warm_up_time = 200


output_dict = control.run_controller(y_ref) #THE ERROR COMES HERE
#

#print(y_ref)

plt.figure(3)
time = np.linspace(0,30,720)

plt.subplot(511)
plt.plot(time,output_dict['y_plot'][:,0]/10**5,color='blue',label='$p_{bh} - ESN$')
plt.plot(time,output_dict['y_pred_plot'][:,0]/10**5,color='red',label='$p_{bh} - True$')
plt.plot(time,y_ref[:,0]/10**5,color='orange', label='$reference$')
#plt.plot(output_dict['y_pred_plot'][:,0],color='orange',label='$p_{wh}$')
plt.grid()
plt.legend()
plt.ylabel('$p_{bh} \ [bar]$')


plt.subplot(512)
plt.plot(time,output_dict['y_plot'][:,1]*3600,color='blue',label='$q - ESN$')
plt.plot(time,y_ref[:,1]*3600,color='orange', label='$reference$')
plt.plot(time,output_dict['y_pred_plot'][:,1]*3600,color='red',label='$q - True$')
plt.grid()
plt.legend()
plt.ylabel('$q$')

plt.subplot(513)
plt.plot(time,output_dict['y_plot'][:,2]/10**5,color='blue',label='$p_{wh} - ESN$')
plt.plot(time,output_dict['y_pred_plot'][:,2]/10**5,color='red',label='$p_{wh} - True$')
plt.grid()
plt.legend()
plt.ylabel('$p_{wh} \ [bar]$')


plt.subplot(514)
plt.step(time,output_dict['u_plot'][:,0],color='red',label='$valve$')
plt.grid()
plt.legend()
plt.ylabel('$\%$')

plt.subplot(515)
plt.step(time,output_dict['u_plot'][:,1],color='red',label='$freq$')
plt.grid()
plt.legend()
plt.ylabel('$freq [Hz]$')
plt.xlabel('$Time [min]$')
#
##sub[0].plot(y_ref)
##sub[0].plot(output_dict['y_plot'][:,0])
##
##
##plt.plot(output_dict['correction_plot'])
##
###sub[1].plot(output_dict['y_plot'][:,1])
###sub[2].plot(output_dict['y_plot'][:,2])
##ax = plt.gca()
##ax.set_xlabel("time")
##sub[0].set_ylabel("CV 1")
##sub[1].set_ylabel("constraint 1")
##sub[2].set_ylabel("constraint 2")
##
##plt.show()
##
print("IAE:",output_dict['IAE']/10**5)